(*
    ############################################################################

    Q[Ĵ߂̕V~[V

    t^C l̉Z

    2003/08/16 쐬

    ############################################################################
*)

unit iQuat;

interface

uses
    Classes, SysUtils, Math, iVector;

type
    TQuaternion = class
    private

    public
        n           : double;   // l
        v           : TVector;  // xNg

        constructor Create(e0, e1, e2, e3: double); overload;
        constructor Create(); overload;

        function    Magnitude(): double;                // 傫
        function    GetVector(): TVector;               // xNgĕԋp
        function    GetScalar(): double;                // XJԋp
                
        function    Plus(q: TQuaternion): TQuaternion;  // ZĎԋp
        function    Minus(q: TQuaternion): TQuaternion; // ZĎԋp
        function    Multiply(s: double): TQuaternion;   // XJςĎԋp
        function    Divide(s: double): TQuaternion;     // XJZĎԋp

        function    Conjugation(): TQuaternion;         // xNgĕԋp
    end;

function    Plus(q1, q2: TQuaternion): TQuaternion;                       // l̉Z
function    Minus(q1, q2: TQuaternion): TQuaternion;                      // ľZ

function    Multiply(q1, q2: TQuaternion): TQuaternion; overload;         // s̏Z
function    Multiply(q: TQuaternion; s: double): TQuaternion; overload;   // XJZ
function    Multiply(s: double; q: TQuaternion): TQuaternion; overload;   // XJZ
function    Multiply(q: TQuaternion; v: TVector): TQuaternion; overload;  // xNgZ
function    Multiply(v: TVector; q: TQuaternion): TQuaternion; overload;  // xNgZ

function    Divide(q: TQuaternion; s: double): TQuaternion;               // XJZ

function    QGetAngle(q: TQuaternion): double;  // xNg̕\Ɋւ]p
function    QGetAxis(q: TQuaternion): TVector;  // xNg̕\ɉPʃxNg
function    QRotate(q1, q2: TQuaternion): TQuaternion;      // lq2q1ɂĉ]
function    QVRotate(q: TQuaternion; v: TVector): TVector;  // xNglɂĉ]

function    MakeQFromEulerAngles(x, y, z: double): TQuaternion; // IC[pl쐬
function    MakeEulerAnglesFromQ(q: TQuaternion): TVector;      // ̔

implementation

// ********************************************************
//
//  TQuaternion
//
// ********************************************************

// ========================================================
constructor TQuaternion.Create(e0, e1, e2, e3: double);
begin
    n := e0;
    v.x := e1;
    v.y := e2;
    v.z := e3;
end;

// ========================================================
constructor TQuaternion.Create();
begin
    Create(0, 0, 0, 0);
end;

// ========================================================
function    TQuaternion.Magnitude(): double;
begin
    Result := Sqrt(n * n + v.x * v.x + v.y * v.y + v.z * v.z);
end;

// ========================================================
function    TQuaternion.GetVector(): TVector;
begin
    Result := TVector.Create(v.x, v.y, v.z);
end;

// ========================================================
function    TQuaternion.GetScalar(): double;
begin
    Result := n;
end;

// ========================================================
function    TQuaternion.Plus(q: TQuaternion): TQuaternion;
begin
    n   := n   + q.n;
    v.x := v.x + q.v.x;
    v.y := v.y + q.v.y;
    v.z := v.z + q.v.z;
    Result := Self;
end;

// ========================================================
function    TQuaternion.Minus(q: TQuaternion): TQuaternion;
begin
    n   := n   - q.n;
    v.x := v.x - q.v.x;
    v.y := v.y - q.v.y;
    v.z := v.z - q.v.z;
    Result := Self;
end;

// ========================================================
function    TQuaternion.Multiply(s: double): TQuaternion;
begin
    n   := n   * s;
    v.x := v.x * s;
    v.y := v.y * s;
    v.z := v.z * s;
    Result := Self;
end;

// ========================================================
function    TQuaternion.Divide(s: double): TQuaternion;
begin
    n   := n   / s;
    v.x := v.x / s;
    v.y := v.y / s;
    v.z := v.z / s;
    Result := Self;
end;

// ========================================================
function    TQuaternion.Conjugation(): TQuaternion;
begin
    Result := TQuaternion.Create(n, - v.x, - v.y, - v.z);
end;

// ********************************************************
//
//  Utility
//
// ********************************************************

// ========================================================
function    Plus(q1, q2: TQuaternion): TQuaternion;
var
    e0, e1, e2, e3: double;
begin
    e0 := q1.n   + q2.n;
    e1 := q1.v.x + q2.v.x;
    e2 := q1.v.y + q2.v.y;
    e3 := q1.v.z + q2.v.z;
    Result := TQuaternion.Create(e0, e1, e2, e3);
end;

// ========================================================
function    Minus(q1, q2: TQuaternion): TQuaternion;
var
    e0, e1, e2, e3: double;
begin
    e0 := q1.n   - q2.n;
    e1 := q1.v.x - q2.v.x;
    e2 := q1.v.y - q2.v.y;
    e3 := q1.v.z - q2.v.z;
    Result := TQuaternion.Create(e0, e1, e2, e3);
end;

// ========================================================
function    Multiply(q1, q2: TQuaternion): TQuaternion;
var
    e0, e1, e2, e3: double;
begin
    e0 := q1.n   * q2.n   - q1.v.x * q2.v.x - q1.v.y * q2.v.y - q1.v.z * q2.v.z;
    e1 := q1.n   * q2.v.x + q1.v.x * q2.n   + q1.v.y * q2.v.z - q1.v.z * q2.v.y;
    e2 := q1.n   * q2.v.y + q1.v.y * q2.n   + q1.v.z * q2.v.x - q1.v.x * q2.v.z;
    e3 := q1.n   * q2.v.z + q1.v.z * q2.n   + q1.v.x * q2.v.y - q1.v.y * q2.v.x;
    Result := TQuaternion.Create(e0, e1, e2, e3);
end;

// ========================================================
function    Multiply(q: TQuaternion; s: double): TQuaternion;
begin
    Result := TQuaternion.Create(q.n, q.v.x, q.v.y, q.v.z);
    Result.Multiply(s);
end;

// ========================================================
function    Multiply(s: double; q: TQuaternion): TQuaternion;
begin
    Result := Multiply(q, s);
end;

// ========================================================
function    Multiply(q: TQuaternion; v: TVector): TQuaternion;
var
    e0, e1, e2, e3: double;
begin
    e0 := - (q.v.x * v.x + q.v.y * v.y + q.v.z * v.z);
    e1 :=    q.n   * v.x + q.v.y * v.z - q.v.z * v.y;
    e2 :=    q.n   * v.y + q.v.z * v.x - q.v.x * v.z;
    e3 :=    q.n   * v.z + q.v.x * v.y - q.v.y * v.x;
    Result := TQuaternion.Create(e0, e1, e2, e3);
end;

// ========================================================
function    Multiply(v: TVector; q: TQuaternion): TQuaternion;
var
    e0, e1, e2, e3: double;
begin
    e0 := - (q.v.x * v.x + q.v.y * v.y + q.v.z * v.z);
    e1 :=    q.n   * v.x + q.v.z * v.y - q.v.y * v.z;
    e2 :=    q.n   * v.y + q.v.x * v.z - q.v.z * v.x;
    e3 :=    q.n   * v.z + q.v.y * v.x - q.v.x * v.y;
    Result := TQuaternion.Create(e0, e1, e2, e3);
end;

// ========================================================

function    Divide(q: TQuaternion; s: double): TQuaternion;
begin
    Result := TQuaternion.Create(q.n, q.v.x, q.v.y, q.v.z);
    Result.Divide(s);
end;

// ========================================================
function    QGetAngle(q: TQuaternion): double;
begin
    Result := (2 * arccos(q.n));
end;

// ========================================================
function    QGetAxis(q: TQuaternion): TVector;
var
    m: double;
begin
    Result := q.GetVector();
    m := Result.Magnitude();
    if m > TOL then Result.Divide(m);
end;

// ========================================================
function    QRotate(q1, q2: TQuaternion): TQuaternion;
var
    c, t: TQuaternion;
begin
    c := q1.Conjugation();
    t := Multiply(q1, q2);
    Result := Multiply(t, c);
    c.Free;
    t.Free;
end;

// ========================================================
function    QVRotate(q: TQuaternion; v: TVector): TVector;
var
    c, t1, t2: TQuaternion;
begin
    c := q.Conjugation();
    t1 := Multiply(q, v);
    t2 := Multiply(t1, c);
    Result := t2.GetVector();
    t2.Free;
    t1.Free;
    c.Free;
end;

// ========================================================
function    MakeQFromEulerAngles(x, y, z: double): TQuaternion;
var
    e0, e1, e2, e3: double;
    roll, pitch, yaw: double;
    croll, cpitch, cyaw, sroll, spitch, syaw: Extended;
    cyawcpitch, syawspitch, cyawspitch, syawcpitch: double;
begin
    roll  := DegToRad(x);
    pitch := DegToRad(y);
    yaw   := DegToRad(z);

    SinCos(0.5 * yaw, syaw, cyaw);
    SinCos(0.5 * pitch, spitch, cpitch);
    SinCos(0.5 * roll, sroll, croll);

    cyawcpitch := cyaw * cpitch;
    syawspitch := syaw * spitch;
    cyawspitch := cyaw * spitch;
    syawcpitch := syaw * cpitch;

    e0 := cyawcpitch * croll + syawspitch * sroll;
    e1 := cyawcpitch * sroll - syawspitch * croll;
    e2 := cyawspitch * croll + syawcpitch * sroll;
    e3 := syawcpitch * croll - cyawspitch * sroll;
    Result := TQuaternion.Create(e0, e1, e2, e3);
end;

// ========================================================
function    MakeEulerAnglesFromQ(q: TQuaternion): TVector;
var
    r11, r21, r31, r32, r33, r12, r13: double;
    q00, q11, q22, q33: double;
    tmp: double;
    x, y, z: double;
begin
    q00 := q.n   * q.n;
    q11 := q.v.x * q.v.x;
    q22 := q.v.y * q.v.y;
    q33 := q.v.z * q.v.z;

    r11 := q00 + q11 - q22 - q33;
    r21 := 2 * (q.v.x * q.v.y + q.n * q.v.z);
    r31 := 2 * (q.v.x * q.v.z - q.n * q.v.y);
    r32 := 2 * (q.v.y * q.v.z + q.n * q.v.x);
    r33 := q00 - q11 - q22 + q33;

    tmp := abs(r31);

    if tmp > 0.999999 then
    begin
        r12 := 2 * (q.v.x * q.v.y - q.n * q.v.z);
        r13 := 2 * (q.v.x * q.v.z + q.n * q.v.y);

        x := RadToDeg(0);                           // [
        y := RadToDeg(- 0.5 * pi * r31 / tmp);      // sb`
        z := RadToDeg(arctan2(- r12, -r31 * r13));  // [
    end else
    begin
        x := RadToDeg(arctan2(r32, r33));           // [
        y := RadToDeg(arcsin(- r31));               // sb`
        z := RadToDeg(arctan2(r21, r11));           // [
    end;

    Result := TVector.Create(x, y, z);
end;

end.
